@@ -10,12 +10,7 @@ import os
 import pytest
 import numpy as np
 
-try:
-    import erfa
-except ModuleNotFoundError:
-    # TODO: This is for backwards compatibility with astropy < 4.2.
-    # When we require 4.2 or greater it should be removed.
-    from astropy import _erfa as erfa
+import erfa
 from astropy.io import fits
 
 from pyuvdata import UVData
@@ -45,6 +40,24 @@ def uvfits_nospw(uvfits_nospw_main):
     return uvfits_nospw_main.copy()
 
 
+@pytest.fixture(scope="session")
+def sma_mir_main():
+    # read in test file for the resampling in time functions
+    uv_object = UVData()
+    testfile = os.path.join(DATA_PATH, "sma_test.mir")
+    uv_object.read(testfile)
+
+    yield uv_object
+
+
+@pytest.fixture(scope="function")
+def sma_mir(sma_mir_main):
+    # read in test file for the resampling in time functions
+    uv_object = sma_mir_main.copy()
+
+    yield uv_object
+
+
 @pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
 @pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
 def test_read_nrao(casa_uvfits):
@@ -75,11 +88,14 @@ def test_break_read_uvfits():
 
 @pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
 @pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
+@pytest.mark.filterwarnings("ignore:UVFITS file is missing AIPS SU table")
 def test_source_group_params(casa_uvfits, tmp_path):
     # make a file with a single source to test that it works
     uv_in = casa_uvfits
-    write_file = str(tmp_path / "outtest_casa.uvfits")
-    write_file2 = str(tmp_path / "outtest_casa2.uvfits")
+    # Writing a source table to UVFITS makes pyuvdata think that the data are
+    # mutli-phase-ctr, so we'll force that in the original file as well
+    write_file = os.path.join(tmp_path, "outtest_casa.uvfits")
+    write_file2 = os.path.join(tmp_path, "outtest_casa2.uvfits")
     uv_in.write_uvfits(write_file)
 
     with fits.open(write_file, memmap=True) as hdu_list:
@@ -126,12 +142,74 @@ def test_source_group_params(casa_uvfits, tmp_path):
 
     uv_out = UVData()
     uv_out.read(write_file2)
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa2.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
 
 @pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
 @pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
-def test_multisource_error(casa_uvfits, tmp_path):
+def test_source_frame_defaults(casa_uvfits, tmp_path):
+    # make a file with a single source to test that it works
+    uv_in = casa_uvfits
+    # Writing a source table to UVFITS makes pyuvdata think that the data are
+    # mutli-phase-ctr, so we'll force that in the original file as well
+    write_file = os.path.join(tmp_path, "outtest_casa.uvfits")
+    write_file2 = os.path.join(tmp_path, "outtest_casa2.uvfits")
+    uv_in.write_uvfits(write_file)
+
+    with fits.open(write_file, memmap=True) as hdu_list:
+        hdunames = uvutils._fits_indexhdus(hdu_list)
+        vis_hdu = hdu_list[0]
+        vis_hdr = vis_hdu.header.copy()
+        raw_data_array = vis_hdu.data.data
+
+        par_names = vis_hdu.data.parnames
+        group_parameter_list = []
+
+        lst_ind = 0
+        for index, name in enumerate(par_names):
+            par_value = vis_hdu.data.par(name)
+            # lst_array needs to be split in 2 parts to get high enough accuracy
+            if name.lower() == "lst":
+                if lst_ind == 0:
+                    # first lst entry, par_value has full lst value
+                    # (astropy adds the 2 values)
+                    lst_array_1 = np.float32(par_value)
+                    lst_array_2 = np.float32(par_value - np.float64(lst_array_1))
+                    par_value = lst_array_1
+                    lst_ind = 1
+                else:
+                    par_value = lst_array_2
+
+            # need to account for PZERO values
+            group_parameter_list.append(par_value - vis_hdr["PZERO" + str(index + 1)])
+
+        vis_hdu = fits.GroupData(
+            raw_data_array, parnames=par_names, pardata=group_parameter_list, bitpix=-32
+        )
+        vis_hdu = fits.GroupsHDU(vis_hdu)
+        vis_hdu.header = vis_hdr
+        del vis_hdu.header["RADESYS"]
+        del vis_hdu.header["EPOCH"]
+        ant_hdu = hdu_list[hdunames["AIPS AN"]]
+
+        hdulist = fits.HDUList(hdus=[vis_hdu, ant_hdu])
+        hdulist.writeto(write_file2, overwrite=True)
+        hdulist.close()
+
+    uv_out = UVData()
+    uv_out.read(write_file2)
+    assert uv_out.phase_center_frame == "icrs"
+
+
+@pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
+@pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
+def test_missing_aips_su_table(casa_uvfits, tmp_path):
     # make a file with multiple sources to test error condition
     uv_in = casa_uvfits
     write_file = str(tmp_path / "outtest_casa.uvfits")
@@ -182,14 +260,20 @@ def test_multisource_error(casa_uvfits, tmp_path):
         hdulist.writeto(write_file2, overwrite=True)
         hdulist.close()
 
-    with pytest.raises(ValueError) as cm:
+    with uvtest.check_warnings(
+        UserWarning,
+        [
+            "UVFITS file is missing AIPS SU table, which is required when ",
+            "Telescope EVLA is not",
+            "The uvw_array does not match the expected values",
+        ],
+    ):
         uv_in.read(write_file2)
-    assert str(cm.value).startswith("This file has multiple sources")
 
 
 @pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
 @pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
-def test_spwsupported():
+def test_multispw_supported():
     """Test reading in a uvfits file with multiple spws."""
     uvobj = UVData()
     testfile = os.path.join(DATA_PATH, "day2_TDEM0003_10s_norx_1scan.uvfits")
@@ -244,7 +328,8 @@ def test_casa_nonascii_bytes_antenna_names():
 
 @pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
 @pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
-def test_readwriteread(tmp_path, casa_uvfits):
+@pytest.mark.parametrize("future_shapes", [True, False])
+def test_readwriteread(tmp_path, casa_uvfits, future_shapes):
     """
     CASA tutorial uvfits loopback test.
 
@@ -252,11 +337,23 @@ def test_readwriteread(tmp_path, casa_uvfits):
     object equality.
     """
     uv_in = casa_uvfits
+
+    if future_shapes:
+        uv_in.use_future_array_shapes()
+
     uv_out = UVData()
     write_file = str(tmp_path / "outtest_casa.uvfits")
 
     uv_in.write_uvfits(write_file)
     uv_out.read(write_file)
+    if future_shapes:
+        uv_out.use_future_array_shapes()
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
     return
@@ -272,6 +369,12 @@ def test_readwriteread_no_lst(tmp_path, casa_uvfits):
     # test that it works with write_lst = False
     uv_in.write_uvfits(write_file, write_lst=False)
     uv_out.read(write_file)
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
     return
@@ -288,6 +391,12 @@ def test_readwriteread_x_orientation(tmp_path, casa_uvfits):
     uv_in.x_orientation = "east"
     uv_in.write_uvfits(write_file)
     uv_out.read(write_file)
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
     return
@@ -306,6 +415,12 @@ def test_readwriteread_antenna_diameters(tmp_path, casa_uvfits):
     )
     uv_in.write_uvfits(write_file)
     uv_out.read(write_file)
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
     return
@@ -335,6 +450,12 @@ def test_readwriteread_large_antnums(tmp_path, casa_uvfits):
     ):
         uv_in.write_uvfits(write_file)
     uv_out.read(write_file)
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
     return
@@ -556,6 +677,11 @@ def test_extra_keywords(casa_uvfits, tmp_path, kwd_names, kwd_values):
     uv_in.write_uvfits(testfile)
     uv_out.read(testfile)
 
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
 
@@ -570,6 +696,12 @@ def test_roundtrip_blt_order(casa_uvfits, tmp_path):
 
     uv_in.write_uvfits(write_file)
     uv_out.read(write_file)
+
+    # make sure filenames are what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
     assert uv_in == uv_out
 
     # test with bda as well (single entry in tuple)
@@ -590,6 +722,7 @@ def test_roundtrip_blt_order(casa_uvfits, tmp_path):
         {"freq_chans": [0]},
         {"polarizations": [-1, -2]},
         {"time_inds": np.array([0, 1])},
+        {"lst_inds": np.array([0, 1])},
         {
             "antenna_nums": np.array([0, 19, 11, 24, 3, 23, 1, 20, 21]),
             "freq_chans": np.arange(12, 22),
@@ -617,6 +750,11 @@ def test_select_read(casa_uvfits, tmp_path, select_kwargs):
         unique_times = np.unique(uvfits_uv2.time_array)
         select_kwargs["time_range"] = unique_times[time_inds]
 
+    if "lst_inds" in select_kwargs.keys():
+        lst_inds = select_kwargs.pop("lst_inds")
+        unique_lsts = np.unique(uvfits_uv2.lst_array)
+        select_kwargs["lst_range"] = unique_lsts[lst_inds]
+
     uvfits_uv.read(casa_tutorial_uvfits, **select_kwargs)
     uvfits_uv2.select(**select_kwargs)
     assert uvfits_uv == uvfits_uv2
@@ -624,6 +762,12 @@ def test_select_read(casa_uvfits, tmp_path, select_kwargs):
     testfile = str(tmp_path / "outtest_casa.uvfits")
     uvfits_uv.write_uvfits(testfile)
     uvfits_uv2.read(testfile)
+
+    # make sure filenames are what we expect
+    assert uvfits_uv.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uvfits_uv2.filename == ["outtest_casa.uvfits"]
+    uvfits_uv.filename = uvfits_uv2.filename
+
     assert uvfits_uv == uvfits_uv2
 
 
@@ -702,6 +846,12 @@ def test_select_read_nospw_pol(casa_uvfits, tmp_path):
     uvfits_uv.read(write_file, polarizations=pols_to_keep)
     uvfits_uv2 = casa_uvfits
     uvfits_uv2.select(polarizations=pols_to_keep)
+
+    # make sure filenames are what we expect
+    assert uvfits_uv.filename == ["outtest_casa.uvfits"]
+    assert uvfits_uv2.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    uvfits_uv.filename = uvfits_uv2.filename
+
     assert uvfits_uv == uvfits_uv2
 
 
@@ -720,6 +870,11 @@ def test_read_uvfits_write_miriad(casa_uvfits, tmp_path):
     uvfits_uv.write_miriad(testfile, clobber=True)
     miriad_uv.read_miriad(testfile)
 
+    # make sure filenames are what we expect
+    assert miriad_uv.filename == ["outtest_miriad"]
+    assert uvfits_uv.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    miriad_uv.filename = uvfits_uv.filename
+
     assert miriad_uv == uvfits_uv
 
     # check that setting the phase_type keyword also works
@@ -736,6 +891,11 @@ def test_read_uvfits_write_miriad(casa_uvfits, tmp_path):
     uvfits_uv.write_miriad(testfile, clobber=True)
     miriad_uv.read_miriad(testfile)
 
+    # make sure filenames are what we expect
+    assert miriad_uv.filename == ["outtest_miriad"]
+    assert uvfits_uv.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    miriad_uv.filename = uvfits_uv.filename
+
     assert miriad_uv == uvfits_uv
 
 
@@ -766,9 +926,32 @@ def test_multi_files(casa_uvfits, tmp_path):
     )
 
     uv1.history = uv_full.history
+
+    # make sure filenames are what we expect
+    assert set(uv1.filename) == {"uv1.uvfits", "uv2.uvfits"}
+    assert uv_full.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    uv1.filename = uv_full.filename
+    uv1._filename.form == (1,)
+
     assert uv1 == uv_full
 
-    # again, setting axis
+
+@pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
+@pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
+def test_multi_files_axis(casa_uvfits, tmp_path):
+    """
+    Reading multiple files at once using "axis" keyword.
+    """
+    uv_full = casa_uvfits
+    testfile1 = str(tmp_path / "uv1.uvfits")
+    testfile2 = str(tmp_path / "uv2.uvfits")
+    uv1 = uv_full.copy()
+    uv2 = uv_full.copy()
+    uv1.select(freq_chans=np.arange(0, 32))
+    uv2.select(freq_chans=np.arange(32, 64))
+    uv1.write_uvfits(testfile1)
+    uv2.write_uvfits(testfile2)
+
     uv1.read([testfile1, testfile2], axis="freq")
     # Check history is correct, before replacing and doing a full object check
     assert uvutils._check_histories(
@@ -780,8 +963,32 @@ def test_multi_files(casa_uvfits, tmp_path):
     )
 
     uv1.history = uv_full.history
+
+    # make sure filenames are what we expect
+    assert set(uv1.filename) == {"uv1.uvfits", "uv2.uvfits"}
+    assert uv_full.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    uv1.filename = uv_full.filename
+    uv1._filename.form == (1,)
+
     assert uv1 == uv_full
 
+
+@pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
+@pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
+def test_multi_files_metadata_only(casa_uvfits, tmp_path):
+    """
+    Reading multiple files at once with metadata only.
+    """
+    uv_full = casa_uvfits
+    testfile1 = str(tmp_path / "uv1.uvfits")
+    testfile2 = str(tmp_path / "uv2.uvfits")
+    uv1 = uv_full.copy()
+    uv2 = uv_full.copy()
+    uv1.select(freq_chans=np.arange(0, 32))
+    uv2.select(freq_chans=np.arange(32, 64))
+    uv1.write_uvfits(testfile1)
+    uv2.write_uvfits(testfile2)
+
     # check with metadata_only
     uv_full = uv_full.copy(metadata_only=True)
     uv1 = UVData()
@@ -797,6 +1004,13 @@ def test_multi_files(casa_uvfits, tmp_path):
     )
 
     uv1.history = uv_full.history
+
+    # make sure filenames are what we expect
+    assert set(uv1.filename) == {"uv1.uvfits", "uv2.uvfits"}
+    assert uv_full.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    uv1.filename = uv_full.filename
+    uv1._filename.form == (1,)
+
     assert uv1 == uv_full
 
 
@@ -829,8 +1043,6 @@ def test_multi_unphase_on_read(casa_uvfits, tmp_path):
             "antenna positions.",
             "Unphasing this UVData object to drift",
             "Unphasing other UVData object to drift",
-            "The uvw_array does not match the expected values given the "
-            "antenna positions.",
         ],
     ):
         uv1.read(np.array([testfile1, testfile2]), unphase_to_drift=True)
@@ -847,6 +1059,13 @@ def test_multi_unphase_on_read(casa_uvfits, tmp_path):
     uv_full.unphase_to_drift()
 
     uv1.history = uv_full.history
+
+    # make sure filenames are what we expect
+    assert set(uv1.filename) == {"uv1.uvfits", "uv2.uvfits"}
+    assert uv_full.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    uv1.filename = uv_full.filename
+    uv1._filename.form = (1,)
+
     assert uv1 == uv_full
 
     # check unphasing when reading only one file
@@ -894,15 +1113,11 @@ def test_multi_phase_on_read(casa_uvfits, tmp_path):
             "Phasing this UVData object to phase_center_radec",
             "The uvw_array does not match the expected values given the "
             "antenna positions.",
-            "The uvw_array does not match the expected values given the "
-            "antenna positions.",
             "Phasing this UVData object to phase_center_radec",
-            "The uvw_array does not match the expected values given the "
-            "antenna positions.",
         ],
     ):
         uv1.read(
-            np.array([testfile1, testfile2]), phase_center_radec=phase_center_radec
+            np.array([testfile1, testfile2]), phase_center_radec=phase_center_radec,
         )
 
     # Check history is correct, before replacing and doing a full object check
@@ -916,6 +1131,13 @@ def test_multi_phase_on_read(casa_uvfits, tmp_path):
 
     uv_full.phase(*phase_center_radec)
     uv1.history = uv_full.history
+
+    # make sure filenames are what we expect
+    assert set(uv1.filename) == {"uv1.uvfits", "uv2.uvfits"}
+    assert uv_full.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    uv1.filename = uv_full.filename
+    uv1._filename.form = (1,)
+
     assert uv1 == uv_full
 
     # check phasing when reading only one file
@@ -951,6 +1173,12 @@ def test_read_ms_write_uvfits_casa_history(tmp_path):
     ms_uv.read_ms(ms_file)
     ms_uv.write_uvfits(testfile, spoof_nonessential=True)
     uvfits_uv.read(testfile)
+
+    # make sure filenames are what we expect
+    assert ms_uv.filename == ["day2_TDEM0003_10s_norx_1src_1spw.ms"]
+    assert uvfits_uv.filename == ["outtest.uvfits"]
+    ms_uv.filename = uvfits_uv.filename
+
     assert ms_uv == uvfits_uv
 
 
@@ -977,3 +1205,56 @@ def test_cotter_telescope_frame(tmp_path):
         ],
     ):
         uvd1.read_uvfits(write_file, read_data=False)
+
+
+@pytest.mark.filterwarnings("ignore:The uvw_array does not match the expected values")
+@pytest.mark.filterwarnings("ignore:Telescope EVLA is not")
+@pytest.mark.parametrize("future_shapes", [True, False])
+def test_readwriteread_reorder_pols(tmp_path, casa_uvfits, future_shapes):
+    """
+    CASA tutorial uvfits loopback test.
+
+    Read in uvfits file, write out new uvfits file, read back in and check for
+    object equality. We check that on-the-fly polarization reordering works.
+    """
+    uv_in = casa_uvfits
+
+    if future_shapes:
+        uv_in.use_future_array_shapes()
+
+    uv_out = UVData()
+    write_file = str(tmp_path / "outtest_casa.uvfits")
+
+    # reorder polarizations
+    polarization_input = uv_in.polarization_array
+    uv_in.reorder_pols(order=[3, 0, 2, 1])
+    assert not np.allclose(uv_in.polarization_array, polarization_input)
+
+    uv_in.write_uvfits(write_file)
+    uv_out.read(write_file)
+    if future_shapes:
+        uv_out.use_future_array_shapes()
+
+    # put polarizations back in order
+    uv_in.reorder_pols(order="AIPS")
+
+    # make sure filename is what we expect
+    assert uv_in.filename == ["day2_TDEM0003_10s_norx_1src_1spw.uvfits"]
+    assert uv_out.filename == ["outtest_casa.uvfits"]
+    uv_in.filename = uv_out.filename
+
+    assert uv_in == uv_out
+
+
+@pytest.mark.parametrize(
+    "freq_val,chan_val,msg",
+    [
+        [-1, 1, "Frequency values must be > 0 for UVFITS!"],
+        [1, 0, "Something is wrong, frequency values not"],
+    ],
+)
+def test_flex_spw_uvfits_write_errs(sma_mir, freq_val, chan_val, msg):
+    sma_mir.freq_array[:] = freq_val
+    sma_mir.channel_width[:] = chan_val
+    with pytest.raises(ValueError, match=msg):
+        sma_mir.write_uvfits("dummy", spoof_nonessential=True)
